/******************************************************************************
*
*  FILENAME :	efacs.c
*
*  DESCRIPTION :
*
*	Source code to front-end of EPAC server, efacs.
*
*  PUBLIC FUNCTIONS :
*
*	int main(int argc, char *argv[])
*
*  NOTES :
*
*	Written by Paul Chesson (1994)
*	E-Mail: chesson@mundil.cs.mu.OZ.AU
*
*/


/******************************************************************************
*  INCLUDE FILES
******************************************************************************/

#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <search.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/stat.h>
#include <sys/types.h>

#include "bugbyte.h"
#include "efacs.h"

#include "file.h"
#include "perms.h"
#include "version.h"


/******************************************************************************
*  PRIVATE DECLARATIONS
******************************************************************************/

/* string comparison macros */

#define STR_EQ3(A, B, C) (STR_EQ((A), (B)) || STR_EQ((A), (C)))

#define IS_A (STR_EQ3(argv[1], "a", "-a"))
#define IS_C (STR_EQ3(argv[1], "c", "-c"))
#define IS_D (STR_EQ3(argv[1], "d", "-d"))
#define IS_I (STR_EQ3(argv[1], "i", "-i"))
#define IS_R (STR_EQ3(argv[1], "r", "-r"))
#define IS_W (STR_EQ3(argv[1], "w", "-w"))


/******************************************************************************
*  PUBLIC FUNCTIONS
******************************************************************************/

/******************************************************************************
*
*  NAME :	int main(int argc, char *argv[])
*
*  DESCRIPTION :
*
*       Carries out the operation specified by the user if access has
*       been authorised.
*
*  INPUTS :
*
*	PARAMETERS:	argc		number of command line arguments
*			argv[]		array of command line arguments
*		
*  OUTPUTS :
*
*	RETURN:		OK		normal termination
*			other		abnornal termination
*
*/

int main(int argc, char *argv[])
{
    char	*filename;		/* file requested by user	*/
    char	names[MAX_NAMES][MAX_NAME_LEN+1];
					/* array of names and aliases..	*/
					/*   that the user satisfies	*/

    /*
     *	    Command Line Processing.
     *	    Exit with usage message if command line syntax incorrect.
     */

    if (argc != 3)
    {
	fprintf(stderr, "%s %s, %s\n", argv[0], EFAC_VERSION, COPYRIGHT);
	fprintf(stderr, "Usage: %s {acdirw} file\n", argv[0]);
	exit(-1);
    }

    filename = argv[2];
    get_names(names);

    /*
     *	    If creating a file, check it does not already exist.
     *	    Otherwise, unless info is requested, check file _does_ exist.
     */

    if (IS_C)
    {
	struct stat	statbuf;

	if (stat(filename, &statbuf) == 0)
	{
	    fprintf(stderr, "%s: %s\n", filename, strerror(EEXIST));
	    exit(EEXIST);
	}
    }
    else if (!IS_I)
    {
	close_file(filename, open_file(filename, "r"));
    }

    /* _____ information _____ */

    if (IS_I)
    {
	int	perms;

	perms = get_perms(filename, FALSE);

	if (perms & I_ACCESS)
	    perms = get_perms(filename, TRUE);

	print_accset(perms, "", TRUE);
    }

    /* _____ append _____ */

    else if (IS_A && (get_perms(filename, FALSE) & A_ACCESS))
	read_file(filename, "a");

    /* _____ copy _____ */

    else if (IS_C && (get_perms(filename, FALSE) & C_ACCESS))
    {
        char	acl_file[FILENAME_MAX+1];
	FILE	*fp;

	/* _____ cannot create .acl files directly _____ */

	if (strlen(filename) >= 4
	    && STR_EQ(filename + strlen(filename) - 4, ".acl"))
	{
	    fprintf(stderr, "%s: %s\n", filename,
	        "Filename cannot have \".acl\" suffix");
	    exit(-1);
	}
	
	/* _____ create default acl file if one doesn't exist _____ */

	strcpy(acl_file, filename);
	strcat(acl_file, ".acl");

	fp = fopen(acl_file, "r");
	if (fp == NULL)
	{
	    fp = open_file(acl_file, "w");
	    fprintf(fp, "%s\ncirw\t%s\n\n", filename,
	        getpwuid(getuid())->pw_name);
	    fprintf(fp, "%s\nirw\t%s\n", acl_file,
	        getpwuid(getuid())->pw_name);
	}
	close_file(acl_file, fp);

	/* _____ read in contents _____ */

	read_file(filename, "w");

    }

    /* _____ delete _____ */

    else if (IS_D && (get_perms(filename, FALSE)  & W_ACCESS))
	delete_file(filename);

    /* _____ read _____ */

    else if (IS_R && (get_perms(filename, FALSE)  & R_ACCESS))
	write_file(filename);

    /* _____ write _____ */

    else if (IS_W && (get_perms(filename, FALSE)  & W_ACCESS))
	read_file(filename, "w");

    /* _____ failed access _____ */

    else
    {
	fprintf(stderr, "%s: %s.\n", filename, strerror(EACCES));
	return EACCES;
    }

    hdestroy();
    return OK;
}

