/******************************************************************************
*
*  FILENAME :	logic.c
*
*  DESCRIPTION :
*
*	Generic package to resolve logical expressions.
*
*  PUBLIC FUNCTIONS :
*
*	int log_eval(const char *exp, const char *err_id,
*		     Boolean (*name_eval)(char *, void *), void *data)
*
*  NOTES :
*
*	Written by Paul Chesson (1994)
*	E-Mail: chesson@mundil.cs.mu.OZ.AU
*
*/


/******************************************************************************
*  INCLUDE FILES
******************************************************************************/

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>

#include "bugbyte.h"
#include "logic.h"


/******************************************************************************
*  PRIVATE DECLARATIONS
******************************************************************************/

/* ----- context ----------------------------------------------------------- */

#define ERROR_LEN 80			/* max length of error messages	*/

typedef struct				/* component of expression	*/
{
	char	*str;			/* pointer to contents		*/
	uint	pos;			/* offset in expression		*/
} Token;

/* ----- data declarations ------------------------------------------------- */

static char	err_msg[ERROR_LEN];	/* string describing last error	*/
static uint	err_tok;		/* token number of last error	*/

/* ----- function prototypes ----------------------------------------------- */

static uint tokenise(const char *exp, Token tokens[]);
static void tok_free(Token tokens[], uint num_tokens);

static int syntax_check(Token tokens[], uint num_tokens);
static void tok_sub(Token tokens[], uint num_tokens,
		    Boolean (*name_eval)(char *, void *), void *data);
static Boolean eval_step(Token tokens[], uint start_token, uint end_token);

static void show_error(const char *err_id, const char *exp, Token tokens[]);

/* ----- macros ----------------------------------------------------------- */

/* _____ character macros _____ */

#define	END_OF_WS_TOKEN(CHAR)						\
    ((CHAR) == TAB || (CHAR) == SPC || (CHAR) == NUL)

/* _____ token classification macros _____ */

#define IS_TRUE(TOK_NUM)	(STR_EQ(tokens[TOK_NUM].str, "t"))
#define IS_FALSE(TOK_NUM)	(STR_EQ(tokens[TOK_NUM].str, "f"))
#define IS_BOOL(TOK_NUM)	(IS_TRUE(TOK_NUM) || IS_FALSE(TOK_NUM))

#define IS_NEGATION(TOK_NUM)	(STR_EQ(tokens[TOK_NUM].str, "!"))
#define IS_CONJUNCT(TOK_NUM)	(STR_EQ(tokens[TOK_NUM].str, "&&"))
#define IS_DISJUNCT(TOK_NUM)	(STR_EQ(tokens[TOK_NUM].str, "||"))
#define IS_LBRACKET(TOK_NUM)	(STR_EQ(tokens[TOK_NUM].str, "("))
#define IS_RBRACKET(TOK_NUM)	(STR_EQ(tokens[TOK_NUM].str, ")"))
#define IS_NAME(TOK_NUM)						\
     ( !IS_NEGATION(TOK_NUM) && !IS_CONJUNCT(TOK_NUM)			\
    && !IS_DISJUNCT(TOK_NUM) && !IS_RBRACKET(TOK_NUM)			\
    && !IS_LBRACKET(TOK_NUM) )

/* ----- token manipulation macro functions ----- */

/*
 *  ADD_TOKEN - Creates the _pos_th logic token whose contents contain
 *  	        the first LEN characters pointed to by START.
 *
 *  NOTES:	Increments pos by length of the token and
 *		increments tok by 1, for the next token.
 */

#define ADD_TOKEN(START, LEN)						\
{									\
    tokens[tok].pos = pos;						\
    pos += (LEN);							\
    MALLOC_TYPE(tokens[tok].str, char, (LEN)+1);			\
    tokens[tok].str[LEN] = NUL;						\
    strncpy(tokens[tok++].str, (START), (LEN));				\
}


/*
 *  DEL_NEXT -	Delete the next NUM_TOKS tokens to the right of
 *		token number _tok_.
 */

#define DEL_NEXT(NUM_TOKS)						\
{									\
    int i;								\
    for (i = tok+1; i <= end_token-(NUM_TOKS); i++)			\
        strcpy(tokens[i].str, tokens[i+(NUM_TOKS)].str);		\
    for (i = end_token-(NUM_TOKS)+1; i <= end_token; i++)		\
        strcpy(tokens[i].str, "");					\
    end_token -= (NUM_TOKS);						\
}

/*
 *  SET_ERROR - Statically store details of last syntax error using
 *		a descriptive message MSG, and the token number
 *		TOK_NUM referenced by the message.
 */

#define SET_ERROR(MSG, TOK_NUM)						\
{									\
    strcpy(err_msg, (MSG));						\
    err_tok = (TOK_NUM);						\
}


/******************************************************************************
*  PUBLIC FUNCTIONS
******************************************************************************/

/******************************************************************************
*
*  NAME :	int log_eval(const char *exp, const char *err_id,
*			     Boolean (*name_eval)(char *, void *), void *data)
*
*  DESCRIPTION :
*
*	Evaluates the boolean value of a logical expression.
*
*  INPUTS :
*
*	PARAMETERS:	exp		logical expression
*			err_id		error identification string for
*					  expression
*			name_eval	function used to evaluate
*					  names in expression
*			data		hook for data required by
*					  name evaluation function
*
*  OUTPUTS :
*
*	RETURN:		FALSE		Expression evaluated false
*			TRUE		Expression evaluated true
*			BLANK		Expression contained no tokens
*			ERROR		Expression failed syntax checking
*
*  NOTES :
*
*	Expression must follow the syntactical rules of C, with the
*	extension of implicit disjunctions between sub-expressions.
*
*	Suggested format for err_id:	"file_name:line_number"
*
*	Format of name evaluation function:
*
*	    Boolean (*name_eval)(char *tok_str, void *data)
*	
*	    tok_str	contents of token string requested for evaluation
*	    data	generic pointer to data structure required by the
*			function -- same as data as passed to log_eval()
*	    returns...	TRUE or FALSE 
*
*/

int log_eval(const char *exp, const char *err_id,
	     Boolean (*name_eval)(char *, void *), void *data)
{
    /*
     *  tokens		array of logical tokens from expression
     *  num_tokens	number of tokens in this array
     *  result		holder for returned value
     */

    Token	*tokens;
    uint 	num_tokens;
    int		result;

    /* _____ resolve blank line cases _____ */

    if (exp == NULL)
	return BLANK;

    MALLOC_TYPE(tokens, Token, strlen(exp)+1);
    if ((num_tokens = tokenise(exp, tokens)) == 0)
    {
	tok_free(tokens, num_tokens);
	return BLANK;
    }

    /* _____ check for syntax errors _____ */

    if (syntax_check(tokens, num_tokens) == ERROR)
    {
	show_error(err_id, exp, tokens);
	tok_free(tokens, num_tokens);
	return ERROR;
    }

    /* _____ evaluate expression _____ */

    tok_sub(tokens, num_tokens, name_eval, data);

    result = eval_step(tokens, 0, num_tokens-1);
    tok_free(tokens, num_tokens);
    return result;

}


/******************************************************************************
*  PRIVATE FUNCTIONS
******************************************************************************/

/******************************************************************************
*
*  NAME :	static uint tokenise(const char *exp, Token tokens[])
*
*  DESCRIPTION :
*
*	Splits expression into logical operator and name tokens.
*
*  INPUTS :
*
*	PARAMETERS:	exp		expression to be tokenised
*		
*  OUTPUTS :
*
*	PARAMETERS:	tokens[]	storage array for created tokens
*		
*	RETURN:		Number of tokens found in expression.
*
*  NOTES :
*
*	!!! tokens[] refer to original expression, so do not destroy it !!!
*
*/

static uint tokenise(const char *exp, Token tokens[])
{
    Token	*ws_tokens;		/* array of whitespace tokens	*/
    uint	num_ws_tokens;		/* number of whitespace tokens	*/
    int		tok, ws_tok;		/* token counters		*/
    char	*cp;

    /*
     *	    Tokenise expression by white space.
     *
     *      strtok() not used since it squeezes the whitespace required
     *      to determine the tokens' positions required for error msgs.
     */

    MALLOC_TYPE(ws_tokens, Token, strlen(exp)+1);
    ws_tok = 0;
    cp = (char *)exp;
    while (cp != NUL)
    {
	while (*cp == SPC || *cp == TAB)
	    cp++;
	if (*cp == NUL)
	    break;
	ws_tokens[ws_tok].str = cp;
	ws_tokens[ws_tok++].pos = cp - exp;
	while (!END_OF_WS_TOKEN(*cp))
	    cp++;
    }
    num_ws_tokens = ws_tok;

    /*
     *      Divide each white space token into
     *      operator tokens (, ), !, ||, &&, and name tokens
     */

    for (ws_tok = 0, tok = 0; ws_tok < num_ws_tokens; ws_tok++)
    {
	char	*sot;			/* start of next name token	*/
	int	pos;			/* offset of next name token	*/

        /* _____ for each character in white space token _____ */

	sot = ws_tokens[ws_tok].str;
	pos = ws_tokens[ws_tok].pos;

	for (cp = ws_tokens[ws_tok].str; !END_OF_WS_TOKEN(*cp); cp++)
	{
	    /* _____ single char symbols _____ */

	    if (*cp == '!' || *cp == '(' || *cp == ')')
	    {
		if (sot != cp)
		    ADD_TOKEN(sot, cp - sot);	/* preceding name token	*/
		ADD_TOKEN(cp, 1);		/* symbol		*/
		sot = cp + 1;
	    }

	    /* _____ double char symbols _____ */

	    if (   (*cp == '&' && *(cp+1) == '&')
		|| (*cp == '|' && *(cp+1) == '|') )
	    {
		if (sot != cp)
		    ADD_TOKEN(sot, cp - sot);	/* preceding name token	*/
		ADD_TOKEN(cp, 2);		/* symbol		*/
		sot = cp + 2;
	    }

	}  /* for each char in white space token */

	if (sot != cp)
	    ADD_TOKEN(sot, cp - sot);		/* final name in ws_token */

    }  /* for each white space token */

    free(ws_tokens);
    return tok;

}


/******************************************************************************
*
*  NAME :	void tok_free(Token tokens[], uint num_tokens)
*
*  DESCRIPTION :
*
*	Frees an array of Token data structures.
*
*  INPUTS :
*
*	PARAMETERS:	tokens[]	tokens to be freed
*			num_tokens	number of tokens in array
*		
*/

static void tok_free(Token tokens[], uint num_tokens)
{
    int i;

    for (i = 0; i < num_tokens; i++)
	free(tokens[i].str);
    free(tokens);
}


/******************************************************************************
*
*  NAME :	static int syntax_check(Token tokens[], uint num_tokens)
*
*  DESCRIPTION :
*
*	Determines if an ordered array of tokens are syntaxically correct.
*
*  INPUTS :
*
*	PARAMETERS:	tokens		token array to check
*			num_tokens	number of tokens in array
*		
*  OUTPUTS :
*
*	RETURN:		OK		syntax is correct
*			ERROR		syntax contains an error
*
*  NOTES :
*
*	Calls SET_ERROR() if a syntax error is found.
*
*/

static int syntax_check(Token tokens[], uint num_tokens)
{
    int lbracket = 0;		/* number of unmatched left brackets	*/
    int i;

    /* _____ syntax ok if there is nothing to check _____ */

    if (num_tokens == 0)
	return OK;

    /*
     *	    Check first token is valid.
     */

    if (IS_RBRACKET(0))
    {
	SET_ERROR("unmatched right parenthesis", 0);
	return ERROR;
    }
    else if (IS_CONJUNCT(0))
    {
	SET_ERROR("missing conjunct", 0);
	return ERROR;
    }
    else if (IS_DISJUNCT(0))
    {
	SET_ERROR("missing disjunct", 0);
	return ERROR;
    }
    if (IS_LBRACKET(0))
	lbracket++;

    /*
     *	    Check inner tokens are succeeded correctly, and that
     *	    right brackets have a matching left bracket
     */

    for (i = 1; i < (num_tokens - 1); i++)
    {
	if (IS_LBRACKET(i) && IS_RBRACKET(i+1))
	{
	    SET_ERROR("empty sub-expression", i);
	    return ERROR;
	}
	else if (IS_LBRACKET(i) && IS_CONJUNCT(i+1))
	{
	    SET_ERROR("missing conjunct", i);
	    return ERROR;
	}
	else if (IS_LBRACKET(i) && IS_DISJUNCT(i+1))
	{
	    SET_ERROR("missing disjunct", i);
	    return ERROR;
	}
	else if ( IS_NEGATION(i) &&
		 (IS_RBRACKET(i+1) || IS_CONJUNCT(i+1) || IS_DISJUNCT(i+1)) )
	{
	    SET_ERROR("illegal operand for negation symbol", i+1);
	    return ERROR;
	}
	else if ( IS_CONJUNCT(i) &&
		 !IS_NAME(i+1) && !IS_LBRACKET(i+1) && !IS_NEGATION(i+1) )
	{
	    SET_ERROR("illegal conjunct", i+1);
	    return ERROR;
	}
	else if ( IS_DISJUNCT(i) &&
		 !IS_NAME(i+1) && !IS_LBRACKET(i+1) && !IS_NEGATION(i+1) )
	{
	    SET_ERROR("illegal disjunct", i+1);
	    return ERROR;
	}
	else if (IS_RBRACKET(i) && --lbracket < 0)
	{
	    SET_ERROR("unmatched right parenthesis", i);
	    return ERROR;
	}
	else if (IS_LBRACKET(i))
	    lbracket++;

    }  /* for each inner token */

    /*
    **	    Ensure last token is valid and brackets balance.
    */

    if (!IS_NAME(num_tokens-1) && !IS_RBRACKET(num_tokens-1))
    {
	SET_ERROR("incomplete expression", num_tokens-1);
	return ERROR;
    }

    if (IS_RBRACKET(num_tokens-1))
	lbracket--;

    /* ----- find unmatched left bracket ----- */

    if (lbracket > 0)
    {
	for (i = num_tokens - 1, lbracket = 0; lbracket <= 0; i--)
	{
	    assert(i >= 0);
	    if (IS_RBRACKET(i))
		lbracket--;
	    else if (IS_LBRACKET(i) && ++lbracket > 0)
	        SET_ERROR("unmatched left parenthesis", i);
	}

	return ERROR;
    }

    return OK;
}


/******************************************************************************
*
*  NAME :	void tok_sub(Token tokens[], uint num_tokens,
*			     boolean (*name_eval)(char *, void *), void *data)
*
*  DESCRIPTION :
*
*	Substitutes name tokens for the boolean valuations "t" or "f",
*
*  INPUTS :
*
*	PARAMETERS:	tokens[]	array of logical tokens
*			num_tokens	number of tokens in array
*			name_eval	function used to evaluate
*					  names in token array
*			data		hook for data required by
*					  name evaluation function
*		
*  OUTPUTS :
*
*	PARAMETERS:	tokens[]	array of logical tokens
*
*  NOTES :
*
*	See log_eval() function notes for details on name_eval().
*
*/

static void tok_sub(Token tokens[], uint num_tokens,
		    Boolean (*name_eval)(char *, void *), void *data)
{
    int i;

    for (i = 0; i < num_tokens; i++)
    {
	if (IS_NAME(i))
	{
	    switch((*name_eval)(tokens[i].str, data))
	    {
		case TRUE:
		    strcpy(tokens[i].str, "t");
		    break;
		case FALSE:
		    strcpy(tokens[i].str, "f");
		    break;
		default:
		    assert(FALSE);
	    }
	}
    }
}


/******************************************************************************
*
*  NAME :	Boolean eval_step(Token tokens[], uint start_token,
*				  uint end_token)
*
*  DESCRIPTION :
*
*	Evaluates a logical sub-expression.
*
*  INPUTS :
*
*	PARAMETERS:	tokens[]	ordered array of logical tokens
*					  after processing by tok_sub
*			start_token	start token number of sub-expression
*			end_token	end token number of sub-expression
*		
*  OUTPUTS :
*
*	PARAMETERS:	tokens[]	token array after sub-expression
*					  evaluation
*		
*	RETURN:		Boolean evaluation of sub-expression
*
*  NOTES :
*
*	This function is recursive, breaking down its input into
*	progressively smaller sub-expressions.
*
*/

static Boolean eval_step(Token tokens[], uint start_token, uint end_token)
{
    uint	tok;			/* current pos'n of token inspection */

    /*
     *	    Evaluate pairs of unnested brackets from left to right.
     *	    Remove brackets after each evaluation.
     */

    for (tok = start_token; tok <= end_token; tok++)
    {
	uint	rbracket;		/* token pos'n of right bracket	*/

	if (IS_RBRACKET(rbracket = tok))
	{
	    while (!IS_LBRACKET(--tok))
		;

	    if (eval_step(tokens, tok+1, rbracket-1) == TRUE)
		strcpy(tokens[tok].str, "t");
	    else
	        strcpy(tokens[tok].str, "f");
	    DEL_NEXT(rbracket-tok);
	}
    }

    /*
     *	    Scan through tokens left-to-right, applying operators
     *	    in precedence.
     */

    /* _____ negation _____ */

    for (tok = start_token; tok <= end_token; tok++)
    {
	if (IS_NEGATION(tok))
	{
	    if (IS_TRUE(tok+1))
		strcpy(tokens[tok].str, "f");
	    else
		strcpy(tokens[tok].str, "t");
	    DEL_NEXT(1);
	}
    }

    /* _____ conjuctions _____ */

    for (tok = start_token; tok < end_token; tok++)
    {
	while (IS_CONJUNCT(tok+1))
	{
	    if (IS_TRUE(tok) && IS_TRUE(tok+2))
		strcpy(tokens[tok].str, "t");
	    else
		strcpy(tokens[tok].str, "f");
	    DEL_NEXT(2);
	}
    }

    /* _____ disjunctions _____ */

    for (tok = start_token; tok < end_token; tok++)
    {
	while ( IS_DISJUNCT(tok+1) || (IS_BOOL(tok) && IS_BOOL(tok+1)) )
	{
	    /* _____ explicit _____ */

	    if (IS_DISJUNCT(tok+1))
	    {
		if (IS_TRUE(tok) || IS_TRUE(tok+2))
		    strcpy(tokens[tok].str, "t");
		else
		    strcpy(tokens[tok].str, "f");
		DEL_NEXT(2);
	    }

	    /* _____ implicit _____ */

	    else
	    {
		if (IS_TRUE(tok) || IS_TRUE(tok+1))
		    strcpy(tokens[tok].str, "t");
		else
		    strcpy(tokens[tok].str, "f");
		DEL_NEXT(1);
	    }

	}  /* while conjunction still exists at current position */

    }  /* scan for conjunctions */

    /*
     *  Base Case - single token
     */

    assert(end_token == start_token);
    assert(IS_BOOL(start_token));
    return IS_TRUE(start_token);

}


/******************************************************************************
*
*  NAME :	void show_error(const char *err_id, const char *exp,
*				Token tokens[])
*
*  DESCRIPTION :
*
*	Display error message for last error set by SET_ERROR().
*
*  INPUTS :
*
*	PARAMETERS:	err_id		error message prefix
*			exp		expression containing error
*			tokens[]	logical tokens from expression
*		
*	GLOBALS:	err_msg		last error message
*			err_tok		token referred to by message
*		
*  NOTES :
*
*	If err_id is the empty string, do not print error.
*	This allows error suppression if required.
*
*/

static void show_error(const char *err_id, const char *exp, Token tokens[])
{
    int		i;

    if (*err_id == NUL)
	return;

    fprintf(stderr, "%s: %s\n\t%s\n\t", err_id, err_msg, exp);

    /*
     *	    Place carets under err_tok
     */

    for (i = 0; i < tokens[err_tok].pos; i++)
    {
	if (exp[i] == '\t')
	    fputc('\t', stderr);
	else
	    fputc(' ', stderr);
    }
	    
    for (i = 0; i < strlen(tokens[err_tok].str); i++)
	fputc('^', stderr);

    fputc('\n', stderr);
}

